//! MAC.cpp

#include <omnetpp.h>

#include "MAC.h"
#include "MessageTypes.h"

#include "debug.h"

Define_Module( MAC );

// Queue sorting function advanced declaration
int compFunc(cObject* a, cObject* b);

//!Do initialization of class data members and any other initialization procedure necessary
void MAC::initialize()
{
    // initialize class data members

	// retrieve module parameters
	epsilon = &par("epsilon");
	NPRIOBITS = (int) par("npriobits");
	
	// initial state
	state = prevState = MAC_STATE_0;

	WATCH(state);
	WATCH(listen);
	WATCH(winner);
	WATCH(prio);
	
	// init message queue
	queue.setName("MAC Queue");
	queue.setup(compFunc, true);
	
	msgRadio = new cMessage("MSG_MAC_TO_RADIO");
	
	sendMsg = failSendMsg = NULL;
	i=0;
	prio=INVALID_PRIO;
	totalTrnmCounter=0;

	observer = (Observer*) simulation.moduleByPath ("observer");
	ASSERT(observer != NULL);

	doMACProtocol(MSG_TYPE_MAC_STATE_ENTER);				
}

//!	Process messages received.
/*!

	\param msg - cMessage* : Message received.
*/
void MAC::handleMessage(cMessage *msg)
{
	if (msg->arrivedOn("from_dg") == true) { 
		// message from data generators, put in the message queue
		ASSERT(msg->kind() == MSG_TYPE_DATA_MSG);
		if (queue.length() < QUEUE_LEN) queue.insert(msg);
		else {
			delete msg;
			return;
		} 

		//std::cout << "Receive msg from DG(" << parentModule()->index() << "):: State=" << state << "\n";

		// if protocol is waiting for a message, signal it
		if (state == MAC_STATE_QUEUEING_ALLOWED) doMACProtocol(MSG_TYPE_MAC_QUEUE);
	} else  {
		// message received from radio or self message: perform MAC protocol
		doMACProtocol(msg->kind());
		if (msg->kind() == MSG_TYPE_DATA_MSG) {
			BUBBLE(this, "%d[%f] Message from remote node", parentModule()->index(), simTime());
		}
		if (msg->isSelfMessage() == false) delete msg;
	}
}

//! Perform MAC Protocol
void MAC::doMACProtocol(int messageType)
{	 	
	DbgOut(dbgLevelTrace, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol: State=%2d; messageType:%2d (i=%2d)\n", parentModule()->index(), simTime(), state, messageType, i);
	
	switch (state) { 
		case MAC_STATE_0: 
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_10 or MAC_STATE_11 
					_radioSetTestMode(); // prepare radio for carrier tx
					_radioSetCarrierSenseOn(); // try to detect a carrier; will always trigger a 1st event signalling the current state
					prio = INVALID_PRIO;
					winner = FALSE;
					_resetTime();					 
					_toState(1);		
					break;
				default: 
					ASSERT(FALSE);
			}
			break;		
//// SYNCRONIZATION									
		case MAC_STATE_1:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_0
					channelIdle=FALSE;
					_setAlarm(SWXRX+TFCS, WAIT_TIME_SWX);																 					
					break;
				case MSG_TYPE_RADIO_END_RX:  
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // channel idle detected
					channelIdle=TRUE;
					if (_isAlarmSet()) break;
					_resetTime();
					_toState(2);
					break;				
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // channel busy detected
					break;
				case MSG_TYPE_MAC_WAIT_TIME_SWX: // 
					if (channelIdle==FALSE) break;
					_resetTime();
					counter=0;
					_toState(2); 																											
					break;
				default:
					ASSERT(FALSE);
			}		
			break;
		case MAC_STATE_2:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_0
					_setAlarm(F, WAIT_TIME_F); // sense the medium for a "long time interval"				
					break;
				case MSG_TYPE_RADIO_END_RX:  
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // channel idle detected
					_cancelAlarm();								
					_resetTime();
					_setAlarm(F, WAIT_TIME_F); // continue to sense the medium for a "long time interval"				
					break;				
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // channel busy detected
					_cancelAlarm();
					_resetTime();		
					_toState(6);					
					break;
				case MSG_TYPE_MAC_WAIT_TIME_F: // ended waiting for a "long time interval" (F)																											
					_resetTime();						
					_toState(3);
					break;
				default:
					ASSERT(FALSE);
			}		
			break;
		case MAC_STATE_3:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_1
					_setAlarm(E, WAIT_TIME_E);																 
					break;							
				case MSG_TYPE_RADIO_CHANNEL_IDLE: //ignore
					break;
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // carrier detected when waiting for E time units, or waiting for someone to start a tournament
					_cancelAlarm();
					_resetTime(); 
					_radioSetCarrierSenseOff(); // stop carrier sensing
					_radioSetCarrierOn(); // CarrierOn								
					_toState(5);
					break;
				case MSG_TYPE_RADIO_END_RX:
					_resetTime();															
					_toState(2);
					break;
				case MSG_TYPE_MAC_QUEUE: // application queued a message
/////////////////CHANGE /START ( adicionei este cdigo )
					if (_isAlarmSet()) break;
					else {
						_resetTime(); 
						_radioSetCarrierSenseOff(); // stop carrier sensing
						_radioSetCarrierOn(); // CarrierOn								
						_toState(4A);
					}
					break;
/////////////////CHANGE /END
				case MSG_TYPE_MAC_WAIT_TIME_E: // signaled at the end of E time units (3->4)
					if (_queueEmpty()==FALSE) {
						_radioSetCarrierSenseOff(); // stop carrier sensing
						_radioSetCarrierOn(); // CarrierOn								
						_toState(4);
					}
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
/////////////////CHANGE /START ( adicionei este cdigo )
		case MAC_STATE_4A:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_2
					_setAlarm(SWXTX+TFCS, WAIT_TIME_SWX); // next state transition +SWX time units
					break;		
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
					// NOP
					break;													
				case MSG_TYPE_MAC_WAIT_TIME_SWX: // signaled at the end of E time units (4->5)
					_resetTime();
					_toState(5); // (4->5)
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
/////////////////CHANGE /END
		case MAC_STATE_4:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_2
					_setAlarm(E+SWXTX, WAIT_TIME_SWX); // next state transition when x=E+SWX (+SWX time units)
					break;		
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
					// NOP
					break;													
				case MSG_TYPE_MAC_WAIT_TIME_SWX: // signaled at the end of E time units (4->5)
					_resetTime();
					_toState(5); // (4->5)
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
		case MAC_STATE_5:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_2
					_setAlarm(H3, WAIT_TIME_P); // next state transition 
					break;		
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
					// NOP
					break;													
				case MSG_TYPE_MAC_WAIT_TIME_P: // signaled at the end of E time units (4->5)
					_radioSetCarrierOff();
					_resetTime();			
					_toState(8); // (4->5)
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
		case MAC_STATE_6:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_2
					_setAlarm(H3-TFCS, WAIT_TIME_P); // next state transition 
					break;		
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // channel idle detected
					_cancelAlarm();
					_resetTime();
					_toState(2);					
					break;								
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
					// NOP
					break;													
				case MSG_TYPE_MAC_WAIT_TIME_P: // signaled at the end of E time units (4->5)
					_toState(7); // (4->5)
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
		case MAC_STATE_7:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_2
					_setAlarm(H3, WAIT_TIME_P); // next state transition 
					break;		
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // ignore					
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
					// NOP
					break;													
				case MSG_TYPE_MAC_WAIT_TIME_P: // signaled at the end of E time units (4->5)
					_resetTime();
					_toState(8);
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;			
		case MAC_STATE_8:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_3
					i=winner_prio=0;
					totalTrnmCounter++;
					if (_queueEmpty()==FALSE) { // we have something to send, so we init for the tournament
						winner=TRUE;
						listen=FALSE;
						_dequeueMsg(); // dequeue highest prio msg; set prio=message prio
						DbgOut(dbgLevelVerbose, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   TRNMT:%2d;PRIO=%2d.\n", parentModule()->index(), simTime(), totalTrnmCounter, prio);
					} else { // we have nothing to send 
						prio = INVALID_PRIO;
						winner=FALSE;
						listen=TRUE;
						sendMsg=NULL;								
						DbgOut(dbgLevelVerbose, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   TRNMT:%2d;PRIO=INV.\n", parentModule()->index(), simTime(), totalTrnmCounter);
					}		
					
					/////////////////////////////////////
					// say to observer that a tournament started 
					observer->tournamentStart();
					/////////////////////////////////////					
						
					_toState(9);
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
//// TOURNAMENT			
		case MAC_STATE_9:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from call _toState()
					_setAlarm(G+(H2+G2)*i, WAIT_TIME_G); // next state transition when x=G+(2H+2G)*i (+G time units)
					
					break;
				// while waiting for the guard time, we ignore other messages from radio
				case MSG_TYPE_RADIO_CHANNEL_BUSY:
				case MSG_TYPE_RADIO_CHANNEL_IDLE:
					// NOP
					break;					
				case MSG_TYPE_MAC_WAIT_TIME_G: // signaled at the end of G time units
					heardDOMBit1=heardDOMBit2=FALSE; 
					if (_prio(i)==0 && winner==TRUE) {
						_winnerPrio(i, 0);
						_radioSetCarrierOn();
					} else {
						_winnerPrio(i, 1);
						_radioSetCarrierSenseOn();
					}
					_toState(10);
					break;																			 											
				default: 
					ASSERT(FALSE);				
			}		
			break;	
		case MAC_STATE_10:
			switch (messageType) {				
				case MSG_TYPE_MAC_STATE_ENTER:
					_setAlarm(H+G+(G2+H2)*i, WAIT_TIME_H); // next state transition when x=H+G+(2G+2H)*i (+H time units)
					break;
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // heard a carrier during the tournament
					//ASSERT((prio_i==1 || winner == FALSE));
					heardDOMBit1=TRUE;
					DbgOut(dbgLevelVerbose, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   W=%2d; hDBit1\n", parentModule()->index(), simTime(), winner);					
					_winnerPrio(i, 0);
					break;
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // ignore
					// NOP
					break;														
				case MSG_TYPE_MAC_WAIT_TIME_H: // ended waiting H time units; check results of this tournament round
					if (_prio(i)==0 && winner==TRUE) {
						_radioSetCarrierOff();
					} else {
						_radioSetCarrierSenseOff();
					}				
				    _toState(11);
					break;
				default:
					ASSERT(FALSE);
			}
			break;
		case MAC_STATE_11:
			switch (messageType) {				
				case MSG_TYPE_MAC_STATE_ENTER: 
					_setAlarm(H+G2+(H2+G2)*i, WAIT_TIME_G); // next state transition when x=H+2G+(2G+2H)*i (+G time units)
					break;
				// while waiting for the guard time, we ignore other messages from radio
				case MSG_TYPE_RADIO_CHANNEL_BUSY:
				case MSG_TYPE_RADIO_CHANNEL_IDLE:
					// NOP
					break;					
				case MSG_TYPE_MAC_WAIT_TIME_G: // signaled at the end of G time units
					if (heardDOMBit1==TRUE) {
						_radioSetCarrierOn();
					} else {
						_radioSetCarrierSenseOn();
					}						
					_toState(12);
					break;																			 											
				default: 
					ASSERT(FALSE);				
			}		
			break;	
		case MAC_STATE_12:
			switch (messageType) {				
				case MSG_TYPE_MAC_STATE_ENTER:
					_setAlarm(H2+G2+(G2+H2)*i, WAIT_TIME_H); // next state transition when x=2H+2G+(2G+2H)*i (+H time units)
					break;
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // heard a carrier during the tournament
					  //ASSERT((prio_i==1 || winner == FALSE));
					  heardDOMBit2=TRUE;
					  _winnerPrio(i, 0);
					break;
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // ignore
					// NOP
					break;														
				case MSG_TYPE_MAC_WAIT_TIME_H: // ended waiting H time units; 
					if (heardDOMBit1==TRUE) {
						_radioSetCarrierOff();
					} else {
						_radioSetCarrierSenseOff();
					}				
					heardDOMBit = (heardDOMBit1 || heardDOMBit2);
					if (winner==TRUE && heardDOMBit==TRUE && _prio(i)==1) {
						winner=FALSE;
					}									
					DbgOut(dbgLevelVerbose, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   W=%2d; hDBit=%2d; PRIO=%2d; PRIO(i)=%2d\n", parentModule()->index(), simTime(), winner, heardDOMBit, prio, _prio(i));
					printf("%2d:[%.5f]\tW=%2d; hDBit=%2d; PRIO=%2d; PRIO(%d)=%2d\n", parentModule()->index(), simTime(), winner, heardDOMBit, prio, i, _prio(i));
			    _toState(13);
					break;
				default:
					ASSERT(FALSE);
			}
			break;		
		case MAC_STATE_13:
			switch (messageType) {				
				case MSG_TYPE_MAC_STATE_ENTER:
					if (i < NPRIOBITS-1) {
						i++;
						_toState(9);
					} else { //  if (i == NPRIOBITS-1)
						if (listen==FALSE) {
							if (winner==TRUE ) {						
								DbgOut(dbgLevelInfo, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   TRNMT %2d:WON\n", parentModule()->index(), simTime(), totalTrnmCounter);
							} else {
								DbgOut(dbgLevelInfo, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   TRNMT %2d:LOST\n", parentModule()->index(), simTime(), totalTrnmCounter);
							}
						} else {
							DbgOut(dbgLevelInfo, dbgMaskMACProtocol, "%2d:[%.5f]\tdoMACProtocol:                                   TRNMT %2d:LIST\n", parentModule()->index(), simTime(), totalTrnmCounter);
						}
						_radioSetNormalMode(); // set radio to normal tx/rx data mode and go to receive state
						_resetTime();															
						_toState(14);					
					}
					break;
				default:
					ASSERT(FALSE);
			}
			break;					
//// Tx/Rx DATA MSG				
		case MAC_STATE_14:
			switch (messageType) {				
				case MSG_TYPE_MAC_STATE_ENTER:
					if (winner==FALSE) {
						if (listen==FALSE) { // lost tournament
//							DbgOut(dbgLevelInfo, dbgMaskMACProtocol, "%d:[%.5f] doMACProtocol: I LOST.\n", parentModule()->index(), simTime());																		
							_enqueueMsg(); // put message back in the message queue; prio = invalid priority
						}				    
						_toState(15);
					} else {
						_setAlarm(H, WAIT_TIME_H); // next message at +G time units
					}
					break;
				case MSG_TYPE_MAC_WAIT_TIME_H: // ended waiting H time units;
					if (_radioSendDataMsg() == SUCCESS) {		///////////////// send packet
						/////////////////////////////////////
						// say to observer that a message was transmitted 
						observer->messageTx(parentModule()->index());
						/////////////////////////////////////						
					} else {
						_failSendMsg();
					}   
					_toState(15); // wait for end Tx message time 
					break;					
			}
			break;				
		case MAC_STATE_15:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER:
					_setAlarm(H+C, WAIT_TIME_C); // next state transition when x=H+C (+C time units)
					break;
				case MSG_TYPE_RADIO_END_TX: // ended sending the message...
					sendMsg = NULL;
					break;
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // ignore
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // ignore
				case MSG_TYPE_RADIO_START_RX: // ignore
				case MSG_TYPE_RADIO_END_RX: // ignore
				case MSG_TYPE_RADIO_RX_ERROR: // ignore
					//NOP 
					break;
				case MSG_TYPE_MAC_WAIT_TIME_C: // ended waiting C time units (Tx message time)
					_resetTime();
					_radioSetTestMode(); 
					_radioSetCarrierSenseOn();
					_toState(16);  
					break;					
				default:
					ASSERT(FALSE);
			}
			break;
		case MAC_STATE_16:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: // enter from (call _toState()) MAC_STATE_0
					_setAlarm(SWXRX+TFCS, WAIT_TIME_SWX);																 					
					break;
				case MSG_TYPE_RADIO_END_RX:  
				case MSG_TYPE_RADIO_CHANNEL_IDLE:
				case MSG_TYPE_RADIO_CHANNEL_BUSY:
					break;
				case MSG_TYPE_MAC_WAIT_TIME_SWX:  
					_resetTime();
					counter++;
					
					/////////////////////////////////////
					// say to observer that a tournament ended 
					if ((winner==FALSE) && (listen==FALSE)) { // lost tournament
					 observer->tournamentEnd(parentModule()->index(), 0);
					} else{
					 if (winner==TRUE) 
					   observer->tournamentEnd(parentModule()->index(), 1);
					  else observer->tournamentEnd(parentModule()->index(), -1);
          } 
					/////////////////////////////////////
					
					if (counter < MAX_TRNMT_COUNTER) {					
						_toState(17);
					} else {
						_toState(0);
					} 																											
					break;
				default:
					ASSERT(FALSE);
			}		
			break;
		case MAC_STATE_17:
			switch (messageType) {
				case MSG_TYPE_MAC_STATE_ENTER: 
					_setAlarm(E+TFCS, WAIT_TIME_E);																 
					break;		
				case MSG_TYPE_RADIO_CHANNEL_IDLE: // ignore
					break;
				case MSG_TYPE_RADIO_CHANNEL_BUSY: // carrier detected 
					_cancelAlarm();
					_resetTime(); 
					_radioSetCarrierSenseOff(); // stop carrier sensing
					_radioSetCarrierOn(); // CarrierOn								
					_toState(5);
					break;
				case MSG_TYPE_MAC_WAIT_TIME_E: // signaled at the end of E time units (3->4)
					_resetTime();
					_toState(3);
					break;
				default: 
					ASSERT(FALSE);				
			}		
			break;
	}
}

//! 
/*!
	\param kind - int : 
*/
bool MAC::sendRadio(int kind)
{
	msgRadio->setKind(kind);
	send((cMessage *)msgRadio->dup(), "to_radio");
	return TRUE;
}

//! 
/*!
	\param msg - cMessage* : 
*/
bool MAC::sendRadio(cMessage *sendMsg)
{
	send(sendMsg, "to_radio");
	sendMsg = NULL;
	return TRUE;
}

//! Set an event to make the MAC protocol develop in a time given by 'wait_time', in seconds
//  This is usually called using macro _waitUntil( _waitTime_clkTks ) because we want times
//  to have as reference the clock "base" time (x) 				
/*!
	\param wait_time - double : wait time in seconds
	\param kind - int : kind of message's event
*/
void MAC::setWait(double wait_time, int kind) 
{
	int wait_clock_ticks = wait_time / CLOCK_TICK_TIME + 1;
	double clock_skew = (double)(*epsilon);
	if (epsilon->type() == 'L' || epsilon->type() == 'D') { // if epsilon is a constant numeric value, just multiply it
		clock_skew *= (wait_clock_ticks-1);
	} else { // otherwise, it may be a distribution; get values from it.
		for (int i=1; i<wait_clock_ticks; i++) clock_skew += (double) (*epsilon);
	}
	clock_skew *= CLOCK_TICK_TIME;
	
//	if (state >= MAC_STATE_25) {
		DbgOut(dbgLevelTrace, dbgMaskMACClock, "%d:[%.5f] MAC: wait_clock_ticks=%d; wait_time=%.5f; clock_skew=%.10f\n", parentModule()->index(), simTime(), wait_clock_ticks, wait_time, clock_skew);
//	}
	
	msgMACAlarm.setKind(kind);
	scheduleAt(simTime() + wait_time + clock_skew, &msgMACAlarm);
}

//! Indicate priority used in the last tournament.
// If, at the time when the last tournament started (or no tournament was yet performed by the node), 
// the node had nothing to send, this will return INVALID_PRIO
int MAC::lastPrio() {
	return prio;
}

//! Indicate if won the last tournament.
bool MAC::lastTournamentWinner() {
	return winner;
} 	

//! Return tournament counter.
int MAC::tournamentCounter() {
	return totalTrnmCounter;
} 	

/*
//! Indicate time of the synchronization in the last tournament.
double MAC::lastSyncTime() {
	return syncTime;
} 		
*/
//! Saves summary results (if any)
void MAC::finish()
{
    // save summary results (if any)
}

MAC::~MAC()
{
	cMessage *msg;
	//queue.clear();
	if (sendMsg != NULL) cancelAndDelete(sendMsg);
	if (failSendMsg != NULL) cancelAndDelete(failSendMsg);
	if (msgRadio != NULL) cancelAndDelete(msgRadio);

	while( ! queue.empty() ) {
		msg = (cMessage *)queue.pop();
		cancelAndDelete(msg);
	}    	   	
}

//! Comparing function
/*!
	\param a - cObject* : Object to compare (a message)
	\param b - cObject* : Object to compare (a message)
	\return according to the message's priorities return: 
		a negative value if p(a)<p(b), 
		0 if p(a)==
		p(b) or 
		a positive value if p(a)>p(b). 
*/
int compFunc(cObject* a, cObject* b) 
{
	cMessage* msga = (cMessage*) a;
	cMessage* msgb = (cMessage*) b;

	return msga->priority() - msgb->priority();
}
