#ifndef __MAC_H__
#define __MAC_H__

#include "Observer.h"

//! MAC simple module class definition
/*

	\author	Nuno Pereira
	\date Act 2005.
*/
class MAC : public cSimpleModule
{
// Member variables
private:
	// retrieved module parameters and pre-calculated data

	//! retrieved module parameter; Clock error
	cPar*		epsilon;

	//! retrieved module parameter; npriobits
	int			NPRIOBITS;
				
	// internal variables
	
	//! MAC protocol self messages, to deliver MAC clock alarms
	cMessage 	msgMACAlarm;
	
	//! MAC protocol messages to deliver to radio
	cMessage 	*msgRadio;
	
	//! Queue of messages to deliver (ordered by message priority)
	cQueue		queue;
	
	// MAC protocol variables

	//! current clock "base" time (see protocol picture...)
	double		x;

	//! MAC protocol current state
	int			state;
	int			prevState;
	
	//! MAC protocol variable: priobits index counter
	int			i;
	
	//!	MAC protocol variable: indicates if the node won the tournament
	bool		winner;
	
	//!	MAC protocol variables
	bool 	heardDOMBit1, heardDOMBit2, heardDOMBit;

	//!	MAC protocol variable: indicates the priority of the message the node is trying to send
	unsigned __int16 prio;
	//!	MAC protocol variable: indicates the priority of the winner 
	unsigned __int8 prio_i;
	
	//!	MAC protocol variable: indicates the priority of the message that won the tournament
	unsigned __int16 winner_prio;
	
	//! Last message dequed (to be/being sent by the protocol)
	cMessage 	*sendMsg;
	cMessage 	*failSendMsg;

	//!	MAC protocol variables
	bool		listen;
	bool 		channelIdle;
	int			counter;

	int			totalTrnmCounter;
	
	//! reference to module used to observe results of the tourmaments 
	Observer *observer;

// Member functions
public:
    Module_Class_Members(MAC,cSimpleModule,0)

	//! Do initialization of class data members and any other initialization procedure necessary
    virtual void initialize();
	//! Process messages received.
    virtual void handleMessage(cMessage *msg);
	//! Saves summary results (if any)
    virtual void finish();

	virtual ~MAC();
	    
// Aditional methods
private:
	void 	doMACProtocol(int messageType);
	bool 	sendRadio(int kind);
	bool 	sendRadio(cMessage *sendMsg);
	void 	setWait(double wait_time, int kind);
	
public:	
	int 	lastPrio();
	bool 	lastTournamentWinner();
  int   tournamentCounter();   	
};	

#define SUCCESS TRUE

//
// This collection of macros is used as an "adaptation layer" between tinyos code
// and simulation code.
//

// macro to get current time in clock ticks (=x in state machine)
#define _currentTime ( (x - simTime()) / CLOCK_TICK_TIME )

// macro to reset current time 
#define _resetTime() {															\
	if (msgMACAlarm.isScheduled()) {											\
		simtime_t  arrival=msgMACAlarm.arrivalTime();							\
		cancelEvent(&msgMACAlarm);												\
		scheduleAt(arrival + (simTime()-x) , &msgMACAlarm);						\
	}																			\
	x = simTime();																\
}

// macro to turn carrier sense on 
#define _radioSetCarrierSenseOn( ) {											\
	sendRadio( MSG_TYPE_RADIO_SET_CARRIER_SENSE );								\
}

// macro to turn carrier sense off 
#define _radioSetCarrierSenseOff() sendRadio( MSG_TYPE_RADIO_SET_IDLE )

// macro to set radio to carrier tx mode
#define _radioSetTestMode() {}

// macro to set radio to receive data mode
#define _radioSetNormalMode() sendRadio( MSG_TYPE_RADIO_SET_RX_DATA )

// macro to set radio to tx data mode
#define _radioSetTx() {} // in simulation, do nothing

// macro to turn carrier pulse on 
#define _radioSetCarrierOn() sendRadio( MSG_TYPE_RADIO_SET_CARRIER_ON )

// macro to turn carrier pulse off 
#define _radioSetCarrierOff() sendRadio( MSG_TYPE_RADIO_SET_CARRIER_OFF )

// macro to send a data message to the radio
#define _radioSendDataMsg() sendRadio( sendMsg )

// macro to give an event when  _currentTime = _when_clkTks 
// time given by '_when_clkTks', in clock ticks, since clock "base time" (x, to match protocol details picture)
// _msgType is only used has a tag to the message kind to deliver 
#define _setAlarm( _when_clkTks, _msgType ) 									\
{																				\
    double _realWaitTime_s = ((x + ((_when_clkTks) * CLOCK_TICK_TIME)) - simTime());	\
	setWait( _realWaitTime_s, MSG_TYPE_MAC_##_msgType );						\
}

// cancel alarm
#define _cancelAlarm() cancelEvent(&msgMACAlarm)

#define _isAlarmSet() (msgMACAlarm.isScheduled())

// macro to call setNextState	
#define _nextState( _next_state ) 												\
{																				\
	prevState = state;															\
	state = MAC_STATE_##_next_state;											\
}		

// macro to transition imediatelly to a state
#define _toState( _next_state ) 												\
{																				\
	prevState = state;															\
	state = MAC_STATE_##_next_state;											\
	doMACProtocol(MSG_TYPE_MAC_STATE_ENTER);									\
}		

// check if send queue is empty 
#define _queueEmpty() (queue.empty())

// dequeue the highest priority message from the queue
#define _dequeueMsg()															\
{																				\
	sendMsg = (cMessage *) queue.head();										\
	ASSERT(sendMsg!=NULL);														\
	prio=sendMsg->priority();													\
	queue.remove(sendMsg);														\
}																	

// put message back in the queue
#define _enqueueMsg() queue.insert(sendMsg)

#define _failSendMsg()

#define _prio( _index ) (!(!(prio & (1 << ((NPRIOBITS - 1) - _index)))))											 
#define _winnerPrio( _index, _val ) { \
	if (_val==1) \
		winner_prio |= (1 << ((NPRIOBITS - 1) - _index)); \
	else { \
		if ((winner_prio & (1 << ((NPRIOBITS - 1) - _index))) != 0) winner_prio ^= (1 << ((NPRIOBITS - 1) - _index));\
	} \
}

#define MAX_NUM_PRIO_BITS	16
#define INVALID_PRIO_INDEX	MAX_NUM_PRIO_BITS
#define INVALID_PRIO 		0xFFFF

#define QUEUE_LEN			10

// MAC protocol states  
enum MAC_STATES {
	MAC_STATE_0,			
	MAC_STATE_1,		
	MAC_STATE_2,
	MAC_STATE_3,
	MAC_STATE_4,
	MAC_STATE_5,
	MAC_STATE_6,
	MAC_STATE_7,
	MAC_STATE_8,
	MAC_STATE_9,
	MAC_STATE_10,
	MAC_STATE_11,		
	MAC_STATE_12,
	MAC_STATE_13,
	MAC_STATE_14,
	MAC_STATE_15,
	MAC_STATE_16,
	MAC_STATE_17,
	MAC_STATE_4A
};

const int MAC_STATE_QUEUEING_ALLOWED = MAC_STATE_3;

const int MAX_TRNMT_COUNTER = 100;

// Clock resolution (s)
const double CLOCK_TICK_TIME	= 0.00003472222; 	//  34,72222 us

// time interval constants (in us)
#define E_us 			690 		// timeout to compensate for innacuracies... 
#define F_us 			44990		// initial idle period
#define H_us 			2390 		// duration of a pulse of a carrier
#define G_us 			1210 		// "guarding" time interval 
#define TFCS_us 		486			// time to detect a carrier (value from experiments)
#define SWXTX_us		192			// switch from idle to TX time (192 us  CC2420 datasheet) 
#define SWXRX_us		320			// switch from idle to RX + first valid CCA time (192 + 128 us  CC2420 datasheet)
#define CMSG_us 		4224		// max time to tx our messages (132 bytes (preamble+SFD+length byte+MPDU+crc)@250Kbps)

// timeouts constants (rounded up to clock ticks)
const unsigned int E 		= ( E_us / 1000000.0 ) / CLOCK_TICK_TIME + 1; 		// 
const unsigned int F 		= ( F_us / 1000000.0 ) / CLOCK_TICK_TIME + 1 ; 		//  	 
const unsigned int H 		= ( H_us / 1000000.0 ) / CLOCK_TICK_TIME + 1; 		// 
const unsigned int H2 		= ( (2*H_us) / 1000000.0 ) / CLOCK_TICK_TIME + 1; 	// 	 
const unsigned int H3 		= ( (3*H_us) / 1000000.0 ) / CLOCK_TICK_TIME + 1; 	// 
const unsigned int H4 		= ( (4*H_us) / 1000000.0 ) / CLOCK_TICK_TIME + 1; 	// 
const unsigned int G 		= ( G_us / 1000000.0 ) / CLOCK_TICK_TIME + 1; 		//  
const unsigned int G2 		= ( (2*G_us) / 1000000.0 ) / CLOCK_TICK_TIME + 1; 	// 
const unsigned int TFCS 	= ( TFCS_us / 1000000.0 ) / CLOCK_TICK_TIME + 1;	// 
const unsigned int SWXRX	= ( SWXRX_us / 1000000.0 ) / CLOCK_TICK_TIME + 1;	// 
const unsigned int SWXTX	= ( SWXTX_us / 1000000.0 ) / CLOCK_TICK_TIME + 1;	//
const unsigned int C		= ( ((CMSG_us) * 1.5) / 1000000.0 ) / CLOCK_TICK_TIME + 1;	// Time to wait for rx/tx messages

#endif
