//! Radio.cpp

#include <omnetpp.h>

#include "MessageTypes.h"
#include "Radio.h"
#include "Node.h"
#include "Observer.h"

#include "debug.h"

//! MAC simple module class definition
/*

	\author	Nuno Pereira
	\date Act 2005.
*/
class Radio : public cSimpleModule
{
// Member variables
private:
	// internal variables
	
	//! radio current state
	int 	state;
	
	//! maintain transmitters
	std::vector <int> transmitters;
	
	//! maintain receive mode (carrier sensing or data)
	int receiveMode;
	
	//! message for self messages
	cMessage *radioMsg;

	//! mac message, duplicated every time we send a message to MAC
	cMessage *macMsg;

	//! reference to module used to observe results of the tourmaments 
	Observer *observer;
		
public:
    Module_Class_Members(Radio,cSimpleModule,0)
	//! Do initialization of class data members and any other initialization procedure necessary
    virtual void initialize();
	//! Process messages received.
    virtual void handleMessage(cMessage *msg);
	//! Saves summary results (if any)
    virtual void finish();

	virtual ~Radio();

// Aditional methods
private:

	void sendMac(int kind, double delay);
	void sendMac(cMessage *msg, double delay);
	void sendAir(int kind);
	void sendAir(cMessage *msg);
	void setWait(int kind, double delay);
	void setCarrierSenseWait(double delay);

	void insertTransmitter(int transmitterId);
	void removeTransmitter(int transmitterId);
	void printTransmitters();
	
};

Define_Module( Radio );

//!Do initialization of class data members and any other initialization procedure necessary
void Radio::initialize()
{
    // initialize class data members
	//state=RADIO_STATE_IDLE;
	state=RADIO_STATE_RX_MODE;
	
	radioMsg = new cMessage("MSG_RADIO");
	macMsg = new cMessage("MSG_RADIO_TO_MAC");
	
	observer = (Observer*) simulation.moduleByPath ("observer");
	ASSERT(observer != NULL);
}

//!	Process messages received.
/*!

	\param msg - cMessage* : Message received.
*/
void Radio::handleMessage(cMessage *msg)
{
	DbgOut(dbgLevelTrace, dbgMaskRadio, "%d:[%.5f] RADIO: State=%d; RM=%d messageType:%d\n", parentModule()->index(), simTime(), state, receiveMode,  msg->kind());
	
	if (msg->arrivedOn("from_air") == true) {	
		// Radio idle or in TX, ignore message
		if (state == RADIO_STATE_IDLE || state == RADIO_STATE_TX || state == RADIO_STATE_TX_MODE) {
//			if (msg->kind()==MSG_TYPE_RADIO_CARRIER_OFF && state == RADIO_STATE_TX) ASSERT(transmitters.size() == 0); // collision?
			
			//! maintain state of heard carrier signals, even if not listenning
			switch (msg->kind()) {
				case MSG_TYPE_RADIO_CARRIER_ON:
					insertTransmitter(msg->senderModuleId());  
					break;
				case MSG_TYPE_RADIO_CARRIER_OFF:
					removeTransmitter(msg->senderModuleId());
					break;
			}
			
			delete msg;
			return;
		}
		
		switch (msg->kind()) {
			case MSG_TYPE_RADIO_CARRIER_ON: 
				if (state == RADIO_STATE_RX_MODE) {
					if (receiveMode == RCV_MODE_CARRIER_SENSE) { 
						// send a message to MAC only after TFCS time
						setWait(MSG_TYPE_RADIO_TFCS, TFCS);
					}
					state=RADIO_STATE_RX;
					insertTransmitter(msg->senderModuleId()); 
				} else if (state == RADIO_STATE_TX) {
					// here we may receive a carrier signal from other nodes
					// but, due to the protocol this should not happen when we are receiving from a node:
//					ASSERT(rxMsg == NULL); // collision ?
				}
				break;
			case MSG_TYPE_RADIO_CARRIER_OFF:
				ASSERT(state==RADIO_STATE_RX || state == RADIO_STATE_RX_MODE); 
				// if TFCS time has not elapsed since change to rx mode,
				// we did not year the carrier long enough to detect it
				cancelEvent(radioMsg);
				removeTransmitter(msg->senderModuleId());
				if (transmitters.size() == 0 && state==RADIO_STATE_RX) {
					sendMac(MSG_TYPE_RADIO_CHANNEL_IDLE, 0);
					state=RADIO_STATE_RX_MODE;
				}
				break;
			case MSG_TYPE_RADIO_START_RX: // (==MSG_TYPE_DATA_MSG)
//				ASSERT(state == RADIO_STATE_RX_MODE || receiveMode != RCV_MODE_DATA); //Collision ? 
				if (!(state == RADIO_STATE_RX_MODE || receiveMode != RCV_MODE_DATA)) {
					
					/////////////////////////////////////
					// say to observer that a collision was detected 
					observer->collision();
					/////////////////////////////////////
					
					sendMac(MSG_TYPE_RADIO_RX_ERROR, 0); // RX_ERROR
					// clear transmitters list
					transmitters.clear();
					state=RADIO_STATE_RX_MODE;
				} else {
					state=RADIO_STATE_RX;
					insertTransmitter(msg->senderModuleId());
					if (receiveMode == RCV_MODE_DATA) {
						ASSERT(transmitters.size() == 1); // something is going wrong...
						sendMac(msg, 0);  							
					} else if (receiveMode == RCV_MODE_CARRIER_SENSE) {
						// send a message to MAC only after TFCS time
						setWait(MSG_TYPE_RADIO_TFCS, TFCS);
					}
				}
				break;
			case MSG_TYPE_RADIO_END_TX: // END_TX received from the air; for us is END_RX
//				ASSERT(state == RADIO_STATE_RX || state == RADIO_STATE_RX_MODE);
				if (state == RADIO_STATE_RX) { 
					removeTransmitter(msg->senderModuleId());
					if (transmitters.size() == 0) {
						state=RADIO_STATE_RX_MODE;
					}
					if (receiveMode == RCV_MODE_DATA) {
						ASSERT(transmitters.size() == 0); // something is going wrong...
						sendMac(MSG_TYPE_RADIO_END_RX, 0); // for us is END_RX
					} else if (receiveMode == RCV_MODE_CARRIER_SENSE) {
							// if TFCS time has not elapsed since change to rx mode,
							// we did not year the carrier long enough to detect it
							cancelEvent(radioMsg); 
							sendMac(MSG_TYPE_RADIO_CHANNEL_IDLE, 0);
					}
				}
				break;		
			default: ASSERT(false);
		}
	} else {
		if (msg->arrivedOn("from_mac") == true) {
			// message from MAC 
			switch (msg->kind()) {
				case MSG_TYPE_RADIO_SET_IDLE:
					ASSERT(state != RADIO_STATE_TX);
					cancelEvent(radioMsg); 
					state = RADIO_STATE_IDLE;
					// clear transmitters
					transmitters.clear();					
					break;
				case MSG_TYPE_RADIO_SET_CARRIER_SENSE:
					ASSERT(state != RADIO_STATE_TX/* && state != RADIO_STATE_RX*/);
					receiveMode = RCV_MODE_CARRIER_SENSE;
					cancelEvent(radioMsg); // just to be safe
					if (state == RADIO_STATE_TX_MODE || state == RADIO_STATE_IDLE) 
						setWait(MSG_TYPE_RADIO_TURN_AROUND_RXTX, TURN_AROUND_RXTX);						
					else {
						state=RADIO_STATE_RX_MODE;
						setWait(MSG_TYPE_RADIO_TFCS, TFCS);
						state=RADIO_STATE_RX_MODE;										
					}
					break;							
				case MSG_TYPE_RADIO_SET_RX_DATA:
//					ASSERT(state != RADIO_STATE_TX);
					cancelEvent(radioMsg); // just to be safe
					receiveMode = RCV_MODE_DATA;
					if (state == RADIO_STATE_TX_MODE || state == RADIO_STATE_IDLE) 
						setWait(MSG_TYPE_RADIO_TURN_AROUND_RXTX, TURN_AROUND_RXTX);						
					else {
						state=RADIO_STATE_RX_MODE;
					}						
					break;
				case MSG_TYPE_RADIO_SET_CARRIER_ON: 
					ASSERT(!(state == RADIO_STATE_RX && receiveMode == RCV_MODE_DATA));
					cancelEvent(radioMsg); // cancel TFCS event, if active
					sendAir(MSG_TYPE_RADIO_CARRIER_ON);
					state=RADIO_STATE_TX;
					// clear transmitters
					transmitters.clear();					
					break;
				case MSG_TYPE_RADIO_SET_CARRIER_OFF:
					ASSERT(state == RADIO_STATE_TX);
					sendAir(MSG_TYPE_RADIO_CARRIER_OFF);
					state=RADIO_STATE_TX_MODE;			
					break;
				case MSG_TYPE_DATA_MSG: // (==MSG_TYPE_RADIO_START_RX)
					//ASSERT(state == RADIO_STATE_TX_MODE || state == RADIO_STATE_RX_MODE || state == RADIO_STATE_IDLE);
					cancelEvent(radioMsg); // cancel TFCS event, if active
					// sendAir(msg) will include the RXTX time, if needed
					sendAir(msg);
					setWait(MSG_TYPE_RADIO_END_TX, FRAME_TIME(msg->length()));		
					state=RADIO_STATE_TX;
					// clear transmitters
					transmitters.clear();
					break;
				default: ASSERT(false);
			}
		} else {
			// self messages
			switch (msg->kind()) {			
				case MSG_TYPE_RADIO_END_TX:
					ASSERT(state == RADIO_STATE_TX);
					sendMac(MSG_TYPE_RADIO_END_TX, 0);
					sendAir(MSG_TYPE_RADIO_END_TX);
					state=RADIO_STATE_TX_MODE;
					break;
				case MSG_TYPE_RADIO_TFCS:
					ASSERT(state==RADIO_STATE_RX || state==RADIO_STATE_RX_MODE);
					// if carrier is on, we should signal a carrier to MAC
					if (receiveMode == RCV_MODE_CARRIER_SENSE) {
						if (transmitters.size() > 0) { 
							sendMac(MSG_TYPE_RADIO_CHANNEL_BUSY, 0);
							state=RADIO_STATE_RX;
						} else {
							sendMac(MSG_TYPE_RADIO_CHANNEL_IDLE, 0);
						}
					}
					// we detected the channel busy. end carrier sensing interval
					break;
				case MSG_TYPE_RADIO_TURN_AROUND_RXTX:
					ASSERT(state != RADIO_STATE_TX);
					state=RADIO_STATE_RX_MODE;
					setWait(MSG_TYPE_RADIO_TFCS, TFCS);					
					break;				
				default: 
					DbgOut(dbgLevelCriticalError, dbgMaskRadio, "%d:[%.5f] MAC: State=%d; messageKind:%d\n", parentModule()->index(), simTime(), state, msg->kind());
					ASSERT(false);		
			}
		}
	} 
	
	if (msg->isSelfMessage() == false) delete msg;
}

void Radio::sendMac(int kind, double delay) {
	macMsg->setKind(kind);
	if (delay>0) sendDelayed((cMessage *)macMsg->dup(), delay, "to_mac"); 
	else send((cMessage *)macMsg->dup(), "to_mac");
}

void Radio::sendMac(cMessage *msg, double delay) {
	if (delay>0) sendDelayed((cMessage *)msg->dup(), delay, "to_mac"); 
	else send((cMessage *)msg->dup(), "to_mac");
}

void Radio::sendAir(int kind) {
	cMessage *msg = new cMessage("MSG_RADIO");
	msg->setKind(kind);
	if (state == RADIO_STATE_TX_MODE || state == RADIO_STATE_TX || state == RADIO_STATE_IDLE) {
		send((cMessage *)msg, "to_air");
	} else sendDelayed((cMessage *)msg, TURN_AROUND_RXTX, "to_air");
}

void Radio::sendAir(cMessage *msg) {
	if (state == RADIO_STATE_TX_MODE || state == RADIO_STATE_TX || state == RADIO_STATE_IDLE) {
		send((cMessage *)msg->dup(), "to_air");
	} else sendDelayed((cMessage *)msg->dup(), TURN_AROUND_RXTX, "to_air");
}

void Radio::setWait(int kind, double delay) {

//	DbgOut(dbgLevelVerbose, dbgMaskRadio, "%d:[%.5f] RADIO: scheduleAt:%.5f messageType:%d\n", parentModule()->index(), simTime(), simTime()+delay, kind);
	if (radioMsg->isScheduled() == FALSE) {
		radioMsg->setKind(kind);		
		scheduleAt(simTime() + delay, radioMsg);
	} else {
		if (radioMsg->kind() != kind) {
			DbgOut(dbgLevelVerbose, dbgMaskRadio, "\n%d:[%.5f] RADIO: *** Radio Message already Scheduled **** \n", parentModule()->index(), simTime());
		}
	}
}

void Radio::insertTransmitter(int transmitterId) {
	std::vector <int>::iterator transmittersIter;

	if (transmitters.size() == 0) { 
		transmitters.insert(transmitters.begin(), transmitterId);
		return;
	}
	
   	for ( transmittersIter = transmitters.begin( ) ; transmittersIter != transmitters.end( ) ; transmittersIter++ )
		if ((*transmittersIter) == transmitterId) break;
	
	if ((*transmittersIter) == transmitterId) return; // do not insert duplicates

	transmitters.insert(transmitters.begin(), transmitterId);  
}

void Radio::removeTransmitter(int transmitterId){
	std::vector <int>::iterator transmittersIter;

	if (transmitters.size() == 0) return;
	
   	for ( transmittersIter = transmitters.begin( ) ; transmittersIter != transmitters.end( ) ; transmittersIter++ )
		if ((*transmittersIter) == transmitterId) break;
	
	if ((*transmittersIter) == transmitterId) {
		transmitters.erase(transmittersIter);
	}

//	DbgOut(dbgLevelTrace, dbgMaskRadio, "\n#%d\n", transmitters.size());
}

void Radio::printTransmitters() {
	std::vector <int>::iterator transmittersIter;

	if (transmitters.size() == 0) return;
	std::cout << "\n";
   	for ( transmittersIter = transmitters.begin( ) ; transmittersIter != transmitters.end( ) ; transmittersIter++ )
   		std::cout << "Tid=" <<(int)(*transmittersIter) << "\n"; 
   		
//	DbgOut(dbgLevelTrace, dbgMaskRadio, "\n#%d\n", transmitters.size());
}

//! Saves summary results (if any)
void Radio::finish()
{	
    // save summary results (if any)
	if (radioMsg != NULL) cancelAndDelete(radioMsg);
	if (macMsg != NULL) cancelAndDelete(macMsg);    
}

Radio::~Radio()
{	
}
